/////////////////////////////////////////////////////////////////////////////
// Name:        src/qt/minifram.cpp
// Purpose:     wxMiniFrame
// Author:      Mariano Reingart
// Copyright:   (c) 2014 wxWidgets dev team
// Licence:     wxWindows licence
/////////////////////////////////////////////////////////////////////////////

// For compilers that support precompilation, includes "wx.h".
#include "wx/wxprec.h"

#if wxUSE_MINIFRAME

#include "wx/minifram.h"

wxIMPLEMENT_DYNAMIC_CLASS(wxMiniFrame, wxFrame);

#endif // wxUSE_MINIFRAME
