/////////////////////////////////////////////////////////////////////////////
// Name:        src/qt/popupwin.cpp
// Author:      Peter Most
// Copyright:   (c) Peter Most
// Licence:     wxWindows licence
/////////////////////////////////////////////////////////////////////////////

// For compilers that support precompilation, includes "wx.h".
#include "wx/wxprec.h"

#include "wx/popupwin.h"

wxPopupWindow::wxPopupWindow()
{
}

wxPopupWindow::wxPopupWindow(wxWindow *WXUNUSED(parent), int WXUNUSED(flags))
{
}

