/*
 *  Name:        wx/qt/defs.h
 *  Author:      Peter Most
 *  Copyright:   (c) Peter Most
 *  Licence:     wxWindows licence
 */

#ifndef _WX_QT_DEFS_H_
#define _WX_QT_DEFS_H_

#ifdef __cplusplus

typedef class QWidget *WXWidget;

#endif

#endif /* _WX_QT_DEFS_H_ */
